@@ -26,9 +26,13 @@
 #define RTC_LAST_EXT_PIN 39
 #define RTC_IS_VALID_EXT_PIN(pin_id) ((1ll << (pin_id)) & RTC_VALID_EXT_PINS)
 
+extern int8_t esp32_rmt_bitstream_channel_id;
+
 extern const mp_obj_type_t esp32_nvs_type;
 extern const mp_obj_type_t esp32_partition_type;
 extern const mp_obj_type_t esp32_rmt_type;
 extern const mp_obj_type_t esp32_ulp_type;
 
+esp_err_t rmt_driver_install_core1(uint8_t channel_id);
+
 #endif // MICROPY_INCLUDED_ESP32_MODESP32_H
